%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%
%
%                            CAEN PLANNER
%
%                               v. 0.2
%
%
%
%   Functionality: Planning with AND/OR graphs
%
%
%   Author: Antoni Ligeza
%
%   File: xplan.pl
%
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%% Knowledge Base %%%%%
% Node types:
% init: initial node; single: single predecessor only; or: logical OR node; and: logical AND node
% pref(node) : prefered nodes declaration; tabu(node) : forbidden nodes declaration
% comp(node) : completed nodes declaration

:- dynamic(pref/1).
:- dynamic(tabu/1).
:- dynamic(comp/1).

prec(z1,init,[]).
prec(z2,init,[]).
prec(z3,init,[]).
prec(z4,init,[]).
prec(z5,init,[]).
prec(z12,single,[z1]).
prec(z8,single,[z3]).
prec(z14,single,[z10]).
prec(z10,single,[z6]).
prec(z15,single,[z11]).
prec(z16,single,[z5]).

prec(z20,or,[z12,z14]).
prec(z17,or,[z12,z13]).
prec(z13,or,[z8,z9]).
prec(z18,or,[z14,z15,z16]).
prec(z11,or,[z6,z7]).

prec(z9,and,[z2,z4]).
prec(z19,and,[z17,z18]).
prec(z6,and,[z3,z4]).
prec(z7,and,[z4,z5]).
prec(z21,and,[z18,z16]).


% xplan([GoalNodes],WorkPlan,SolutionPlan)
% A plan is generated if all GoalNodes nodes are completed
xplan(GoalNodes,CompletedNodes,WorkPlan,SolutionPlan):-
	completed(GoalNodes),!,
	SolutionPlan=WorkPlan,
        union(CompletedNodes,GoalNodes,AllPlanNodes),
	output(AllPlanNodes,SolutionPlan).
xplan(GoalNodes,CompletedNodes,WorkPlan,SolutionPlan):-
	select_best(Node,GoalNodes,LessGoalNodes),             % select a best node to be developed (may be redefined)
	process(Node,Type,NewNodes,CompletedNodes),                                % develop the selected node - find its predecessors
	add(NewNodes,Type,LessGoalNodes,NewGoalNodes),              % add new predecessors to goal nodes
	assign(Node,Type,NewNodes,CompletedNodes,WorkPlan,NewWorkplan),                     % add the Node to current plan
	xplan(NewGoalNodes,CompletedNodes,NewWorkplan,SolutionPlan).          % go on with depth-first search

select_best(Node,GoalNodes,LessGoalNodes):- select(Node,GoalNodes,LessGoalNodes),!.

process(Node,_,[],CompletedNodes):-  member(Node,CompletedNodes),!.
process(Node,Type,[]):- Type=init, prec(Node,Type,[]),!.
process(Node,Type,NewNodes):- Type=single, prec(Node,Type,NewNodes),!.
process(Node,Type,NewNodes):- Type=and, prec(Node,Type,NewNodes),!.
process(Node,Type,NewNodes):-
	Type=or,
	prec(Node,Type,Nodes),
	member(NewNode,Nodes),
	NewNodes=[NewNode]. % member selects the node to be explored in the case of an OR node; can be redefined.

add(_,init,LessGoalNodes,NewGoalNodes):- NewGoalNodes=LessGoalNodes,!.
add(NewNodes,_,LessGoalNodes,NewGoalNodes):- union(NewNodes,LessGoalNodes,NewGoalNodes),!.

assign(Node,_,_,CompletedNodes,WorkPlan,NewWorkplan):- member(Node,CompletedNodes),NewWorkplan=WorkPlan,!.
assign(Node,Type,NewNodes,WorkPlan,NewWorkplan):- union([prec(Node,Type,NewNodes)],WorkPlan,NewWorkplan).



completed([]):- !.
completed([Z|T]):- prec(Z,init,[]),!, completed(T). % any initial node is completed
completed([Z|T]):- comp(Z),!, completed(T).         % any completed node is also completed


output(Nodes,Plan):-
	write('***** PLAN *****'),nl,
	write('Plan nodes: '), write(Nodes),nl,
	write('Plan: '), write(Plan),nl,nl.

